@@ -1,6 +1,7 @@
 from .trezor import TrezorClient
 
+
 class KeepkeyClient(TrezorClient):
-    def __init__(self, path, password='', expert=False):
+    def __init__(self, path, password="", expert=False):
         super(KeepkeyClient, self).__init__(path, password, expert)
-        self.type = 'Keepkey'
\ No newline at end of file
+        self.type = "Keepkey"
