@@ -3,6 +3,7 @@ from io import StringIO
 from werkzeug.wrappers import Response
 from datetime import datetime
 from numbers import Number
+from math import isnan
 from ..util.tx import decoderawtransaction
 from ..util.price_providers import get_price_at
 
@@ -59,6 +60,30 @@ def wallets_overview():
     )
 
 
+################## Failed wallets fix #######################
+
+
+@wallets_endpoint.route("/failed_wallets/", methods=["POST"])
+@login_required
+def failed_wallets():
+    if request.method == "POST":
+        action = request.form["action"]
+        if action == "retry_loading_wallets":
+            app.specter.wallet_manager.update()
+        elif action == "delete_failed_wallet":
+            try:
+                wallet = json.loads(request.form["wallet_data"])
+                fullpath = wallet["fullpath"]
+                delete_file(fullpath)
+                delete_file(fullpath + ".bkp")
+                delete_file(fullpath.replace(".json", "_addr.csv"))
+                delete_file(fullpath.replace(".json", "_txs.csv"))
+                app.specter.wallet_manager.update()
+            except Exception as e:
+                flash(f"Failed to delete failed wallet: {str(e)}", "error")
+    return redirect("/")
+
+
 ################## New wallet #######################
 
 
@@ -272,7 +297,6 @@ def new_wallet(wallet_type):
 
             return render_template(
                 "wallet/new_wallet/new_wallet_keys.jinja",
-                purposes=purposes,
                 cosigners=devices,
                 wallet_type=wallet_type,
                 sigs_total=len(devices),
@@ -297,7 +321,6 @@ def new_wallet(wallet_type):
                 ]
                 return render_template(
                     "wallet/new_wallet/new_wallet_keys.jinja",
-                    purposes=purposes,
                     cosigners=devices,
                     wallet_type=wallet_type,
                     sigs_total=len(devices),
@@ -332,7 +355,6 @@ def new_wallet(wallet_type):
                 ]
                 return render_template(
                     "wallet/new_wallet/new_wallet_keys.jinja",
-                    purposes=purposes,
                     cosigners=devices,
                     wallet_type=wallet_type,
                     sigs_total=len(devices),
@@ -351,7 +373,6 @@ def new_wallet(wallet_type):
                 err = "Failed to create wallet..."
                 return render_template(
                     "wallet/new_wallet/new_wallet_keys.jinja",
-                    purposes=purposes,
                     cosigners=cosigners,
                     wallet_type=wallet_type,
                     sigs_total=len(devices),
@@ -399,7 +420,6 @@ def new_wallet(wallet_type):
         if action == "preselected_device":
             return render_template(
                 "wallet/new_wallet/new_wallet_keys.jinja",
-                purposes=purposes,
                 cosigners=[
                     app.specter.device_manager.get_by_alias(request.form["device"])
                 ],
@@ -438,7 +458,7 @@ def wallet(wallet_alias):
 
 
 ###### Wallet transaction history ######
-@wallets_endpoint.route("/wallet/<wallet_alias>/history/")
+@wallets_endpoint.route("/wallet/<wallet_alias>/history/", methods=["GET", "POST"])
 @login_required
 def history(wallet_alias):
     try:
@@ -446,6 +466,15 @@ def history(wallet_alias):
     except SpecterError as se:
         app.logger.error("SpecterError while wallet_tx: %s" % se)
         return render_template("base.jinja", error=se, specter=app.specter, rand=rand)
+
+    tx_list_type = "txlist"
+
+    if request.method == "POST":
+        action = request.form["action"]
+        if action == "freezeutxo":
+            wallet.toggle_freeze_utxo(request.form.getlist("selected_utxo"))
+            tx_list_type = "utxo"
+
     # update balances in the wallet
     app.specter.check_blockheight()
     wallet.get_balance()
@@ -455,6 +484,7 @@ def history(wallet_alias):
         "wallet/history/wallet_history.jinja",
         wallet_alias=wallet_alias,
         wallet=wallet,
+        tx_list_type=tx_list_type,
         specter=app.specter,
         rand=rand,
     )
@@ -525,22 +555,38 @@ def send_new(wallet_alias):
     addresses = [""]
     labels = [""]
     amounts = [0]
-    fee_rate = 0.0
+    amount_units = ["btc"]
     err = None
     ui_option = "ui"
     recipients_txt = ""
+    subtract = False
+    subtract_from = 1
+    fee_options = "dynamic"
+    fee_rate = 0.0
+    fee_rate_blocks = 6
+    rbf = True
+    selected_coins = request.form.getlist("coinselect")
     if request.method == "POST":
-        action = request.form["action"]
+        action = request.form.get("action")
         if action == "createpsbt":
             i = 0
             addresses = []
             labels = []
             amounts = []
-            ui_option = request.form.get("ui_option")
+            amount_units = []
+            ui_option = request.form.get("ui_option", "ui")
             if "ui" in ui_option:
                 while "address_{}".format(i) in request.form:
                     addresses.append(request.form["address_{}".format(i)])
-                    amounts.append(float(request.form["btc_amount_{}".format(i)]))
+                    amount = 0.0
+                    try:
+                        amount = float(request.form["btc_amount_{}".format(i)])
+                    except ValueError:
+                        pass
+                    if isnan(amount):
+                        amount = 0.0
+                    amounts.append(amount)
+                    amount_units.append(request.form["amount_unit_{}".format(i)])
                     labels.append(request.form["label_{}".format(i)])
                     if request.form["label_{}".format(i)] != "":
                         wallet.setlabel(addresses[i], labels[i])
@@ -553,6 +599,7 @@ def send_new(wallet_alias):
                         amounts.append(float(output.split(",")[1].strip()) / 1e8)
                     else:
                         amounts.append(float(output.split(",")[1].strip()))
+                    labels.append("")
             addresses = [
                 address.lower()
                 if address.startswith(("BC1", "TB1", "BCRT1"))
@@ -560,25 +607,28 @@ def send_new(wallet_alias):
                 for address in addresses
             ]
             subtract = bool(request.form.get("subtract", False))
-            subtract_from = int(request.form.get("subtract_from", 1)) - 1
+            subtract_from = int(request.form.get("subtract_from", 1))
+            fee_options = request.form.get("fee_options")
+            if fee_options:
+                if "dynamic" in fee_options:
+                    fee_rate = float(request.form.get("fee_rate_dynamic"))
+                    fee_rate_blocks = int(request.form.get("fee_rate_blocks"))
+                else:
+                    if request.form.get("fee_rate"):
+                        fee_rate = float(request.form.get("fee_rate"))
             rbf = bool(request.form.get("rbf", False))
             selected_coins = request.form.getlist("coinselect")
             app.logger.info("selected coins: {}".format(selected_coins))
-            if "dynamic" in request.form.get("fee_options"):
-                fee_rate = float(request.form.get("fee_rate_dynamic"))
-            else:
-                if request.form.get("fee_rate"):
-                    fee_rate = float(request.form.get("fee_rate"))
             try:
                 psbt = wallet.createpsbt(
                     addresses,
                     amounts,
                     subtract=subtract,
-                    subtract_from=subtract_from,
+                    subtract_from=subtract_from - 1,
                     fee_rate=fee_rate,
+                    rbf=rbf,
                     selected_coins=selected_coins,
                     readonly="estimate_fee" in request.form,
-                    rbf=rbf,
                 )
                 if psbt is None:
                     err = "Probably you don't have enough funds, or something else..."
@@ -593,7 +643,7 @@ def send_new(wallet_alias):
                 app.logger.error(e)
             if err is None:
                 if "estimate_fee" in request.form:
-                    return psbt
+                    return jsonify(success=True, psbt=psbt)
                 return render_template(
                     "wallet/send/sign/wallet_send_sign_psbt.jinja",
                     psbt=psbt,
@@ -603,42 +653,9 @@ def send_new(wallet_alias):
                     specter=app.specter,
                     rand=rand,
                 )
-        elif action == "importpsbt":
-            try:
-                b64psbt = "".join(request.form["rawpsbt"].split())
-                psbt = wallet.importpsbt(b64psbt)
-            except Exception as e:
-                flash("Could not import PSBT: %s" % e, "error")
-                return redirect(
-                    url_for("wallets_endpoint.import_psbt", wallet_alias=wallet_alias)
-                )
-            return render_template(
-                "wallet/send/sign/wallet_send_sign_psbt.jinja",
-                psbt=psbt,
-                labels=labels,
-                wallet_alias=wallet_alias,
-                wallet=wallet,
-                specter=app.specter,
-                rand=rand,
-            )
-        elif action == "openpsbt":
-            psbt = ast.literal_eval(request.form["pending_psbt"])
-            return render_template(
-                "wallet/send/sign/wallet_send_sign_psbt.jinja",
-                psbt=psbt,
-                labels=labels,
-                wallet_alias=wallet_alias,
-                wallet=wallet,
-                specter=app.specter,
-                rand=rand,
-            )
-        elif action == "deletepsbt":
-            try:
-                wallet.delete_pending_psbt(
-                    ast.literal_eval(request.form["pending_psbt"])["tx"]["txid"]
-                )
-            except Exception as e:
-                flash("Could not delete Pending PSBT!", "error")
+            else:
+                if "estimate_fee" in request.form:
+                    return jsonify(success=False, error=str(err))
         elif action == "rbf":
             try:
                 rbf_tx_id = request.form["rbf_tx_id"]
@@ -691,12 +708,28 @@ def send_new(wallet_alias):
                 specter=app.specter,
                 rand=rand,
             )
+    show_advanced_settings = (
+        ui_option != "ui"
+        or subtract
+        or fee_options != "dynamic"
+        or fee_rate_blocks != 6
+        or not rbf
+        or selected_coins
+    )
     return render_template(
         "wallet/send/new/wallet_send.jinja",
         psbt=psbt,
         ui_option=ui_option,
         recipients_txt=recipients_txt,
-        labels=labels,
+        recipients=list(zip(addresses, amounts, amount_units, labels)),
+        subtract=subtract,
+        subtract_from=subtract_from,
+        fee_options=fee_options,
+        fee_rate=fee_rate,
+        fee_rate_blocks=fee_rate_blocks,
+        rbf=rbf,
+        selected_coins=selected_coins,
+        show_advanced_settings=show_advanced_settings,
         wallet_alias=wallet_alias,
         wallet=wallet,
         specter=app.specter,
@@ -723,6 +756,16 @@ def send_pending(wallet_alias):
             except Exception as e:
                 app.logger.error("Could not delete Pending PSBT: %s" % e)
                 flash("Could not delete Pending PSBT!", "error")
+        elif action == "openpsbt":
+            psbt = ast.literal_eval(request.form["pending_psbt"])
+            return render_template(
+                "wallet/send/sign/wallet_send_sign_psbt.jinja",
+                psbt=psbt,
+                wallet_alias=wallet_alias,
+                wallet=wallet,
+                specter=app.specter,
+                rand=rand,
+            )
     pending_psbts = wallet.pending_psbts
     ######## Migration to multiple recipients format ###############
     for psbt in pending_psbts:
@@ -739,7 +782,7 @@ def send_pending(wallet_alias):
     )
 
 
-@wallets_endpoint.route("/wallet/<wallet_alias>/send/import")
+@wallets_endpoint.route("/wallet/<wallet_alias>/send/import", methods=["GET", "POST"])
 @login_required
 def import_psbt(wallet_alias):
     try:
@@ -747,6 +790,25 @@ def import_psbt(wallet_alias):
     except SpecterError as se:
         app.logger.error("SpecterError while wallet_send: %s" % se)
         return render_template("base.jinja", error=se, specter=app.specter, rand=rand)
+    if request.method == "POST":
+        action = request.form["action"]
+        if action == "importpsbt":
+            try:
+                b64psbt = "".join(request.form["rawpsbt"].split())
+                psbt = wallet.importpsbt(b64psbt)
+            except Exception as e:
+                flash("Could not import PSBT: %s" % e, "error")
+                return redirect(
+                    url_for("wallets_endpoint.import_psbt", wallet_alias=wallet_alias)
+                )
+            return render_template(
+                "wallet/send/sign/wallet_send_sign_psbt.jinja",
+                psbt=psbt,
+                wallet_alias=wallet_alias,
+                wallet=wallet,
+                specter=app.specter,
+                rand=rand,
+            )
     err = None
     return render_template(
         "wallet/send/import/wallet_importpsbt.jinja",
@@ -969,6 +1031,7 @@ def broadcast(wallet_alias):
 
 @wallets_endpoint.route("/wallet/<wallet_alias>/decoderawtx/", methods=["GET", "POST"])
 @login_required
+@app.csrf.exempt
 def decoderawtx(wallet_alias):
     try:
         wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
@@ -998,6 +1061,7 @@ def decoderawtx(wallet_alias):
     "/wallet/<wallet_alias>/rescan_progress", methods=["GET", "POST"]
 )
 @login_required
+@app.csrf.exempt
 def rescan_progress(wallet_alias):
     try:
         wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
@@ -1043,6 +1107,7 @@ def set_label(wallet_alias):
 
 @wallets_endpoint.route("/wallet/<wallet_alias>/txlist", methods=["POST"])
 @login_required
+@app.csrf.exempt
 def txlist(wallet_alias):
     try:
         wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
@@ -1069,6 +1134,7 @@ def txlist(wallet_alias):
 
 @wallets_endpoint.route("/wallet/<wallet_alias>/utxo_list", methods=["POST"])
 @login_required
+@app.csrf.exempt
 def utxo_list(wallet_alias):
     try:
         wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
@@ -1077,7 +1143,7 @@ def utxo_list(wallet_alias):
         search = request.form.get("search", None)
         sortby = request.form.get("sortby", None)
         sortdir = request.form.get("sortdir", "asc")
-        txlist = wallet.utxo
+        txlist = wallet.full_utxo
         for tx in txlist:
             if not tx.get("label", None):
                 tx["label"] = wallet.getlabel(tx["address"])
@@ -1091,6 +1157,7 @@ def utxo_list(wallet_alias):
 
 @wallets_endpoint.route("/wallets_overview/txlist", methods=["POST"])
 @login_required
+@app.csrf.exempt
 def wallets_overview_txlist():
     try:
         idx = int(request.form.get("idx", 0))
@@ -1116,6 +1183,7 @@ def wallets_overview_txlist():
 
 @wallets_endpoint.route("/wallets_overview/utxo_list", methods=["POST"])
 @login_required
+@app.csrf.exempt
 def wallets_overview_utxo_list():
     try:
         idx = int(request.form.get("idx", 0))
@@ -1135,6 +1203,7 @@ def wallets_overview_utxo_list():
 
 @wallets_endpoint.route("/wallet/<wallet_alias>/addresses_list/", methods=["POST"])
 @login_required
+@app.csrf.exempt
 def addresses_list(wallet_alias):
     """Return a JSON with keys:
         addressesList: list of addresses with the properties
@@ -1174,6 +1243,7 @@ def addresses_list(wallet_alias):
 
 @wallets_endpoint.route("/wallet/<wallet_alias>/addressinfo/", methods=["POST"])
 @login_required
+@app.csrf.exempt
 def addressinfo(wallet_alias):
     try:
         wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
@@ -1303,7 +1373,7 @@ def utxo_csv(wallet_alias):
         sortdir = request.args.get("sortdir", "desc")
         txlist = json.loads(
             process_txlist(
-                wallet.utxo,
+                wallet.full_utxo,
                 idx=0,
                 limit=0,
                 search=search,
@@ -1541,7 +1611,7 @@ def addresses_list_to_csv(wallet):
         else:
             balance_on_address = 0
             if address_info.used:
-                for tx in wallet.utxo:
+                for tx in wallet.full_utxo:
                     if tx.get("address", "") == address:
                         balance_on_address += tx.get("amount", 0)
         row += (balance_on_address,)
