@@ -35,7 +35,15 @@ addrtypes = {
 
 class WalletManager:
     # chain is required to manage wallets when bitcoind is not running
-    def __init__(self, data_folder, rpc, chain, device_manager, path="specter"):
+    def __init__(
+        self,
+        bitcoin_core_version_raw,
+        data_folder,
+        rpc,
+        chain,
+        device_manager,
+        path="specter",
+    ):
         self.data_folder = data_folder
         self.chain = chain
         self.rpc = rpc
@@ -43,6 +51,8 @@ class WalletManager:
         self.device_manager = device_manager
         self.is_loading = False
         self.wallets = {}
+        self.failed_load_wallets = []
+        self.bitcoin_core_version_raw = bitcoin_core_version_raw
         self.update(data_folder, rpc, chain)
 
     def update(self, data_folder=None, rpc=None, chain=None):
@@ -71,6 +81,7 @@ class WalletManager:
         existing_names = list(self.wallets.keys())
         # list of wallet to keep
         keep_wallets = []
+        self.failed_load_wallets = []
         try:
             if self.working_folder is not None and self.rpc is not None:
                 wallets_files = load_jsons(self.working_folder, key="name")
@@ -84,9 +95,13 @@ class WalletManager:
                             self.rpc.loadwallet(
                                 os.path.join(self.rpc_path, wallet_alias)
                             )
-                            wallets[wallet_name] = Wallet.from_json(
+                            loaded_wallet = Wallet.from_json(
                                 wallets_files[wallet], self.device_manager, self
                             )
+                            if loaded_wallet:
+                                wallets[wallet_name] = loaded_wallet
+                            else:
+                                raise Exception("Failed to load wallet")
                             # Lock UTXO of pending PSBTs
                             if len(wallets[wallet_name].pending_psbts) > 0:
                                 for psbt in wallets[wallet_name].pending_psbts:
@@ -105,27 +120,62 @@ class WalletManager:
                                             ].pending_psbts[psbt]["tx"]["vin"]
                                         ],
                                     )
+                            if len(wallets[wallet_name].frozen_utxo) > 0:
+                                wallets[wallet_name].rpc.lockunspent(
+                                    False,
+                                    [
+                                        {
+                                            "txid": utxo.split(":")[0],
+                                            "vout": int(utxo.split(":")[1]),
+                                        }
+                                        for utxo in wallets[wallet_name].frozen_utxo
+                                    ],
+                                )
                         except RpcError as e:
-                            logger.warn(
+                            logger.warning(
                                 f"Couldn't load wallet {wallet_alias} into core.\
 Silently ignored! RPC error: {e}"
                             )
+                            self.failed_load_wallets.append(
+                                {
+                                    **wallets_files[wallet],
+                                    "loading_error": str(e).replace("'", ""),
+                                }
+                            )
                         except Exception as e:
-                            logger.warn(
+                            logger.warning(
                                 f"Couldn't load wallet {wallet_alias}.\
 Silently ignored! Wallet error: {e}"
                             )
+                            self.failed_load_wallets.append(
+                                {
+                                    **wallets_files[wallet],
+                                    "loading_error": str(e).replace("'", ""),
+                                }
+                            )
                     else:
                         if wallet_name not in existing_names:
                             # ok wallet is already there
                             # we only need to update
                             try:
-                                wallets[wallet_name] = Wallet.from_json(
+                                loaded_wallet = Wallet.from_json(
                                     wallets_files[wallet], self.device_manager, self
                                 )
+                                if loaded_wallet:
+                                    wallets[wallet_name] = loaded_wallet
+                                else:
+                                    raise Exception("Failed to load wallet")
                             except Exception as e:
-                                logger.warn(f"Failed to load wallet {wallet_name}: {e}")
-                                logger.warn(traceback.format_exc())
+                                logger.warning(
+                                    f"Failed to load wallet {wallet_name}: {e}"
+                                )
+                                logger.warning(traceback.format_exc())
+                                self.failed_load_wallets.append(
+                                    {
+                                        **wallets_files[wallet],
+                                        "loading_error": str(e).replace("'", ""),
+                                    }
+                                )
                         else:
                             # wallet is loaded and should stay
                             keep_wallets.append(wallet_name)
@@ -146,7 +196,7 @@ Silently ignored! Wallet error: {e}"
 
     def get_by_alias(self, alias):
         for wallet_name in self.wallets:
-            if self.wallets[wallet_name].alias == alias:
+            if self.wallets[wallet_name] and self.wallets[wallet_name].alias == alias:
                 return self.wallets[wallet_name]
         raise SpecterError("Wallet %s does not exist!" % alias)
 
@@ -189,8 +239,13 @@ Silently ignored! Wallet error: {e}"
             change_descriptor = "%s(%s)" % (el, change_descriptor)
         recv_descriptor = AddChecksum(recv_descriptor)
         change_descriptor = AddChecksum(change_descriptor)
-
-        self.rpc.createwallet(os.path.join(self.rpc_path, wallet_alias), True)
+        if self.bitcoin_core_version_raw >= 210000:
+            # Use descriptor wallet
+            self.rpc.createwallet(
+                os.path.join(self.rpc_path, wallet_alias), True, True, "", False, True
+            )
+        else:
+            self.rpc.createwallet(os.path.join(self.rpc_path, wallet_alias), True)
 
         w = Wallet(
             name,
@@ -211,16 +266,20 @@ Silently ignored! Wallet error: {e}"
             devices,
             sigs_required,
             {},
+            [],
             os.path.join(self.working_folder, "%s.json" % wallet_alias),
             self.device_manager,
             self,
         )
         # save wallet file to disk
-        if self.working_folder is not None:
+        if w and self.working_folder is not None:
             w.save_to_file()
         # get Wallet class instance
-        self.wallets[name] = w
-        return w
+        if w:
+            self.wallets[name] = w
+            return w
+        else:
+            raise ("Failed to create new wallet")
 
     def delete_wallet(
         self, wallet, bitcoin_datadir=get_default_datadir(), chain="main"
@@ -252,14 +311,25 @@ Silently ignored! Wallet error: {e}"
             wallet.save_to_file()
         self.update()
 
-    def full_txlist(self, page, validate_merkle_proofs=False):
+    def full_txlist(
+        self,
+        fetch_transactions=True,
+        validate_merkle_proofs=False,
+        current_blockheight=None,
+    ):
+        """Returns a list of all transactions in all wallets loaded in the wallet_manager.
+        #Parameters:
+        #    fetch_transactions (bool): Update the TxList CSV caching by fetching transactions from the Bitcoin RPC
+        #    validate_merkle_proofs (bool): Return transactions with validated_blockhash
+        #    current_blockheight (int): Current blockheight for calculating confirmations number (None will fetch the block count from the RPC)
+        """
         txlists = [
             [
                 {**tx, "wallet_alias": wallet.alias}
                 for tx in wallet.txlist(
-                    page,
-                    limit=(100 // len(self.wallets)),
+                    fetch_transactions=fetch_transactions,
                     validate_merkle_proofs=validate_merkle_proofs,
+                    current_blockheight=current_blockheight,
                 )
             ]
             for wallet in self.wallets.values()
@@ -270,6 +340,25 @@ Silently ignored! Wallet error: {e}"
                 result.append(tx)
         return list(reversed(sorted(result, key=lambda tx: tx["time"])))
 
+    def full_utxo(self):
+        """Returns a list of all UTXOs in all wallets loaded in the wallet_manager."""
+        txlists = [
+            [
+                {
+                    **utxo,
+                    "label": wallet.getlabel(utxo["address"]),
+                    "wallet_alias": wallet.alias,
+                }
+                for utxo in wallet.full_utxo
+            ]
+            for wallet in self.wallets.values()
+        ]
+        result = []
+        for txlist in txlists:
+            for tx in txlist:
+                result.append(tx)
+        return list(reversed(sorted(result, key=lambda tx: tx["time"])))
+
     def delete(self, specter):
         """Deletes all the wallets"""
         for w in self.wallets:
